@@ -88,6 +88,12 @@ static int decode_arg (int * i, int argc, char ** argv)
 				rf |= EXT2_APPEND_FL;
 				rem = 1;
 				break;
+#endif
+#ifdef EXT2_NOATIME_FL
+			case 'A':
+				rf |= EXT2_NOATIME_FL;
+				rem = 1;
+				break;
 #endif
 			case 'c':
 				rf |= EXT2_COMPR_FL;
@@ -114,9 +120,9 @@ static int decode_arg (int * i, int argc, char ** argv)
 				rem = 1;
 				break;
 			case 'v':
+				(*i)++;
 				if (*i >= argc)
 					usage ();
-				(*i)++;
 				version = strtol (argv[*i], &tmp, 0);
 				if (*tmp)
 				{
@@ -144,6 +150,11 @@ static int decode_arg (int * i, int argc, char ** argv)
 			case 'a':
 				af |= EXT2_APPEND_FL;
 				break;
+#endif
+#ifdef EXT2_NOATIME_FL
+			case 'A':
+				af |= EXT2_NOATIME_FL;
+				break;
 #endif
 			case 'c':
 				af |= EXT2_COMPR_FL;
@@ -180,6 +191,11 @@ static int decode_arg (int * i, int argc, char ** argv)
 			case 'a':
 				sf |= EXT2_APPEND_FL;
 				break;
+#endif
+#ifdef EXT2_NOATIME_FL
+			case 'A':
+				sf |= EXT2_NOATIME_FL;
+				break;
 #endif
 			case 'c':
 				sf |= EXT2_COMPR_FL;
@@ -310,9 +326,9 @@ void main (int argc, char ** argv)
 		fprintf (stderr, "= is incompatible with - and +\n");
 		exit (1);
 	}
-	if (!(add || rem || set))
+	if (!(add || rem || set || set_version))
 	{
-		fprintf (stderr, "Must use =, - or +\n");
+		fprintf (stderr, "Must use '-v', =, - or +\n");
 		exit (1);
 	}
 	for (j = i; j < argc; j++)
